!*********************************************************************************!
!  _____________________________________________________________________________  !
! |                                                                             | !
! |   Filename: PairDiag.f90                                Public Version 1.0  | !
! |                                                                             | !  
! |     Mon, Mar 09th, 2020                                                     | !
! |     Author: X.Y. Liu                                                        | !
! |                                                                             | !
! |_____________________________________________________________________________| !
!  _____________________________________________________________________________  !
! |                                                                             | !
! |   PairDiag: A program to solve the first eigenvalue and the corresponding   | !
! |             eigenvector of general pairing Hamiltonians.                    | !
! |                                                                             | ! 
! |_____________________________________________________________________________| !
! |                                                                             | !
! |   There is a predefined derived data type in the module, type Diag_Par.     | !
! |                                                                             | !
! |     Users needs to declare a variable of this type in their program.        | !
! |     This variable will be a mandatory parameter of the calculation.         | !
! |     Please use the provided Main.f90 in the archive as a reference.         | !
! |                                                                             | !
! |_____________________________________________________________________________| !
! |                                                                             | !
! |   Custom Input Parameters in the variable of type Diag_Par:                 | !
! |                                                                             | !
! |     Diag_Par%Orbit: Total Pair number available.                            | !
! |     Diag_Par%Pairs: Occupied numbers of Pairs.                              | !
! |     Diag_Par%SPE:   1-D Array of Single Particle Energy.                    | !
! |     Diag_Par%P_F:   2-D Array of Pairing Force.                             | !
! |                                                                             | !
! |_____________________________________________________________________________| !
! |                                                                             | !
! |   Optional Parameters:                                                      | !
! |                                                                             | !
! |     Lanc_Limit:       Total Pair number available.                          | !
! |     Lanc_Error:       Occupied numbers of Pairs.                            | !
! |     Dimension_Limits: 1-D Array of Single Particle Energy.                  | !
! |     Print_Mode:       2-D Array of Pairing Force.                           | !
! |                                                                             | !
! |_____________________________________________________________________________| !
! |                                                                             | !
! |   Public subroutine for Even-mass or Odd-mass System:                       | !
! |                                                                             | !
! |     call Diag_Solver(Diag_Par, [0]): Restart Lanczos.                       | !
! |     call Diag_Solver(Diag_Par, 1):   Lanczos without restart.               | !
! |                                                                             | !
! |_____________________________________________________________________________| !
! |                                                                             | !
! |   Public subroutine only for Odd-mass System (X = 0 or 1):                  | !
! |                                                                             | !
! |     call Diag_Solver(Diag_Par, X, [0]): All subspace calculation.           | !
! |     call Diag_Solver(Diag_Par, X, N):   The Nth subspace calculation.       | !
! |                                                                             | !
! |_____________________________________________________________________________| !
! |                                                                             | !
! |   Output Results in the variable of type Diag_Par:                          | !
! |                                                                             | !
! |     Diag_Par%Energy_Ground: Ground State Energy.                            | !
! |     Diag_Par%Monopole_Min:  Minimum Diagonal Element.                       | !
! |     Diag_Par%N_Occup:       1-D Array of Occupation Numbers.                | !
! |                                                                             | !
! |_____________________________________________________________________________| !
!_________________________________________________________________________________!


module PairDiag 

  use omp_lib
  implicit none

  private
  public Diag_Par
  public Diag_Solver

  !Optional Parameters:
  integer(kind=8):: Lanc_Limit = 50
  real(kind=8):: Lanc_Error = 1D-5
  integer(kind=8):: Dimension_Limits = 1D+9
  integer(kind=1):: Print_Mode = 0

  !Predefined derived data type:
  type Diag_Par
    integer(kind=8):: Orbit
    real(kind=8):: Pairs
    real(kind=8),dimension(63):: SPE
    real(kind=8),dimension(63, 63):: P_F
    real(kind=8):: Energy_Ground
    real(kind=8):: Monopole_Min
    real(kind=8),dimension(63):: N_Occup
  end type

  !Public subroutine:
  interface Diag_Solver
    module procedure Diag_Solver0
    module procedure Diag_Solver1
    module procedure Diag_Solver2
  end interface

  !Variables
  integer(kind=8):: B_Dimension, L_Dimension
  integer(kind=8):: N_Orbit, Posit_Min, Total, Occup
  real(kind=8):: N_Pairs, Energy_Ground, Energy_Blocked
  real(kind=8):: Monopole_Min, Monopole_Max, Monopole_Trun
  integer(kind=1):: Convergence, Truncated, Run_Mode, Block_Mode

  !Arrays
  real(kind=8),dimension(5):: Eigens
  real(kind=8),dimension(63, 63):: P_F
  real(kind=8),dimension(63):: SPE, SPM, N_Occup
  integer(kind=8),allocatable,dimension(:, :):: C_Array
  real(kind=8),allocatable,dimension(:):: I_Vector, Q_Vector
  integer(kind=1),allocatable,dimension(:, :):: O_Array, V_Array
  integer(kind=8),allocatable,dimension(:):: B_Array, Monopole_Hist 
  real(kind=8),allocatable,dimension(:, :):: Q_Matrix, L_Matrix, T_Matrix, P_Matrix

contains

  subroutine Diag_Solver0(DiagPar)
    implicit none
    type(Diag_Par),intent(inout):: DiagPar
    N_Orbit = DiagPar%Orbit
    N_Pairs = DiagPar%Pairs
    SPE = DiagPar%SPE
    P_F = DiagPar%P_F
    Run_Mode = 0
    Block_Mode = 0
    call Diag_Analysis()
    DiagPar%Energy_Ground = Energy_Ground
    DiagPar%Monopole_Min = Monopole_Min
    DiagPar%N_Occup = N_Occup
    return
  end subroutine Diag_Solver0

  subroutine Diag_Solver1(DiagPar, Mode)
    implicit none
    type(Diag_Par),intent(inout):: DiagPar
    integer(kind=4),intent(in):: Mode
    N_Orbit = DiagPar%Orbit
    N_Pairs = DiagPar%Pairs
    SPE = DiagPar%SPE
    P_F = DiagPar%P_F
    Run_Mode = Mode
    Block_Mode = 0
    call Diag_Analysis()
    DiagPar%Energy_Ground = Energy_Ground
    DiagPar%Monopole_Min = Monopole_Min
    DiagPar%N_Occup = N_Occup
    return
  end subroutine Diag_Solver1

  subroutine Diag_Solver2(DiagPar, Mode, Block)
    implicit none
    type(Diag_Par),intent(inout):: DiagPar
    integer(kind=4),intent(in):: Mode, Block
    N_Orbit = DiagPar%Orbit
    N_Pairs = DiagPar%Pairs
    SPE = DiagPar%SPE
    P_F = DiagPar%P_F
    Run_Mode = Mode
    Block_Mode = Block
    call Diag_Analysis()
    DiagPar%Energy_Ground = Energy_Ground
    DiagPar%Monopole_Min = Monopole_Min
    DiagPar%N_Occup = N_Occup
    return
  end subroutine Diag_Solver2

  subroutine Diag_Analysis()
    implicit none
    if(N_Orbit>63 .or. N_Orbit<1) then
      write(*, "(1X,A)") "N_Orbit is expected in [1, 63]!"
      stop
    end if
    if(N_Pairs>N_Orbit .or. N_Pairs<1) then
      write(*, "(1X,A)") "N_Pairs is expected in [1, N_Orbit]!"
      stop
    end if
    do Total=1, N_Orbit
      do Occup=Total+1, N_Orbit
        if(P_F(Total, Occup)/=P_F(Occup, Total)) then
          write(*, "(1X,A)") "P_F should be a real symmetric matrix!"
          stop
        end if
      end do
    end do
    Total = N_Orbit
    Occup = int(N_Pairs)
    if(N_Pairs-Occup/=0 .and. N_Pairs-Occup/=0.5) then
      write(*, "(1X,A)") "N_Pairs should be an integer or half-integer!"
      stop
    end if
    if(Total<Occup) then
      write(*, "(1X,A)") "Total<Occup is not accepted!"
      stop
    end if
    if(N_Pairs-Occup==0) then
      call Even_System()
    else
      if(Block_Mode>Total .or. Block_Mode<0) then
        write(*, "(1X,A)") "Block is out the range of [1, Total]!"
        stop
      end if
      call Odds_System()
    end if
    return
  end subroutine Diag_Analysis

  subroutine Even_System()
    implicit none
    Energy_Blocked = 0
    L_Dimension = Lanc_Limit
    call Basis_Initialize()
    call Lanczos_Restart()
    call Results_Output()
    call Destory()
    return
  end subroutine Even_System

  subroutine Odds_System()
    implicit none
    real(kind=8),dimension(5):: Eigens_B
    real(kind=8),dimension(63, 63):: P_F_B
    integer(kind=8):: Total_B, i, j, k, l, m, B
    real(kind=8),dimension(63):: SPE_B, N_Occup_B
    real(kind=8):: Energy_Ground_B, Monopole_Min_B
    SPE_B = SPE
    P_F_B = P_F
    Total_B = Total
    Total = Total_B - 1
    Energy_Ground_B = 1D+9
    do i=1, Total_B
      if(Block_Mode>0.and.i/=Block_Mode) cycle
      l = 0
      do j=1, Total_B
        if(j==i) cycle
        l = l + 1
        SPE(l) = SPE_B(j)
        m = 0
        do k=1, Total_B
          if(k==i) cycle
          m = m + 1
          P_F(l, m) = P_F_B(j, k)
        end do
      end do
      Energy_Blocked = SPE_B(i)
      L_Dimension = Lanc_Limit
      write(*, *)
      write(*, "(1X,A11,I4,A)") "Blocking : ", i, "   ............."
      call Basis_Initialize()
      call Lanczos_Restart()
      call Results_Output()
      if(Energy_Ground_B>Energy_Ground) then
        Energy_Ground_B = Energy_Ground
        Monopole_Min_B = Monopole_Min
        k = 0
        B = i
        do j=1, Total
          if(j==i) then
            N_Occup_B(j) = 1.0000
            k = k + 1
          end if
          k = k + 1
          N_Occup_B(k) = N_Occup(j)
        end do
        Eigens_B = Eigens
      end if
      call Destory()
    end do
    Eigens = Eigens_B
    N_Occup = N_Occup_B
    Energy_Ground = Energy_Ground_B
    Monopole_Min = Monopole_Min_B
    write(*, *)
    write(*, "(1X,A17,I4,A)") "Final Blocking : ", B, "   ............."
    write(*, *)
    write(*, "(1X,A)") "Results..."
    write(*, "(3X,A17,F22.14)") "Energy_Ground   :", Energy_Ground
    write(*, "(3X,A17,F22.14)") "Monopole_Min    :", Monopole_Min
    write(*, *)
    if(Run_Mode/=0) then
      write(*, "(1X,A)") "First Five Eigenvalues..."
      do i=1, min(5, L_Dimension)
        write(*, "(3X,A11,I1,A3,F22.14)") "Eigenvalue ", i, " : ", Eigens(i)
      end do
      write(*, *)
    end if
    SPE = SPE_B
    P_F = P_F_B
    Total = Total_B
    return
  end subroutine Odds_System

  subroutine Lanczos_Restart()
    implicit none
    call Vector_Initialize()
    call Lanczos_Iteration()
    call QR_Decompose()
    do while(Convergence==0)
      call Vector_Restart()
      call Lanczos_Iteration()
      call QR_Decompose()
    end do
  end subroutine Lanczos_Restart

  function Monopole_E(State)
    implicit none
    integer(kind=8),intent(in):: State
    real(kind=8):: Monopole_E
    integer(kind=1):: i
    Monopole_E = Energy_Blocked
    do i=1, Total
      if(btest(State, i-1)) Monopole_E = Monopole_E + SPM(i)
    end do
    return
  end function Monopole_E

  subroutine Basis_Initialize()
    implicit none
    integer(kind=8):: State, State_I, State_F, Dimension_Sum
    integer(kind=4):: NT
    do State=1, Total
      SPM(State) = 2.0*SPE(State) + P_F(State, State)
    end do
    do State_I=1, Total-1
      Monopole_Min = SPM(State_I)
      do State_F=State_I+1, Total
        if(SPM(State_F)<Monopole_Min) then
          SPM(State_I) = SPM(State_F)
          SPM(State_F) = Monopole_Min
          Monopole_Min = SPM(State_I)
        end if
      end do
    end do
    B_Dimension = Combin_Num(Total, Occup)
    State_I = 0
    State_F = 0
    Monopole_Min = Energy_Blocked
    Monopole_Max = Energy_Blocked
    do State=0, Occup-1
      State_I = ibset(State_I, State)
      State_F = ibset(State_F, State+Total-Occup)
      Monopole_Min = Monopole_Min + SPM(State+1)
      Monopole_Max = Monopole_Max + SPM(State+Total-Occup+1)
    end do
    do State=1, Total
      SPM(State) = 2.0*SPE(State) + P_F(State, State)
    end do
    Truncated = 0
    Monopole_Trun = Monopole_Max
    Dimension_Sum = B_Dimension
    if(B_Dimension>Dimension_Limits) then
      allocate(Monopole_Hist(100000000))
      Truncated = 1
      Monopole_Hist = 0
      Dimension_Sum = 0
      State = State_I
      do while(State<=State_F)
        NT = int((99999999)*(Monopole_E(State)-Monopole_Min)/(Monopole_Max-Monopole_Min)) + 1
        Monopole_Hist(NT) = Monopole_Hist(NT) + 1
        call Next_State(State)
      end do
      do NT=1, 100000000
        Dimension_Sum = Dimension_Sum + Monopole_Hist(NT)
        if(Dimension_Sum>=Dimension_Limits.and.Monopole_Trun==Monopole_Max) then
          Monopole_Trun = (Monopole_Max-Monopole_Min)/100000000*(NT-1)+Monopole_Min
        end if
      end do
      deallocate(Monopole_Hist)
      State = State_I
      B_Dimension = 0
      do while(State<=State_F)
        if(Trun_State(State)) B_Dimension = B_Dimension + 1
        call Next_State(State)
      end do
    end if
    allocate(B_Array(B_Dimension))
    if(Print_Mode==0) then
      write(*, *)
      write(*, "(1X,A)") "Information..."
      write(*, "(3X,A33,F12.4)") "Truncation Energy :              ", Monopole_Trun
      write(*, "(3X,A33,I12)")   "Dimension before the truncation :", Dimension_Sum
      write(*, "(3X,A33,I12)")   "Dimension after the truncation  :", B_Dimension
      if(Truncated==0) then
        write(*, "(3X,A27)")   "The space is not truncated!"
      else
        write(*, "(3X,A27)")   "The space is truncated!!!!!"
      end if
      write(*, "(3X,A9,I3,A9,I3,A8)") "There are", Occup, "pairs in", Total, "orbits!"
      write(*, *)
    end if
    Posit_Min = 0
    State = State_I
    B_Dimension = 0
    do while(State<=State_F)
      if(Trun_State(State)) then
        B_Dimension = B_Dimension + 1
        B_Array(B_Dimension) = State
        if(Monopole_Min==Monopole_E(State)) Posit_Min = B_Dimension
      end if
      call Next_State(State)
    end do
    if(Posit_Min==0) then
      write(*, "(1X,A)") "This cannot happen!!!"
      stop
    end if
    NT = 1
    !$OMP PARALLEL
    !$ NT = OMP_GET_NUM_THREADS()
    !$OMP END PARALLEL
    allocate(O_Array(Occup, NT))
    allocate(V_Array(Total-Occup, NT))
    allocate(C_Array(Total, Occup))
    do State_I=1, Total
      do State_F=1, Occup
        C_Array(State_I, State_F) = Combin_Num(State_I-1, State_F)
      end do
    end do
    if(L_Dimension>B_Dimension) L_Dimension = B_Dimension
    allocate(I_Vector(B_Dimension))
    allocate(Q_Matrix(B_Dimension, L_Dimension))
    allocate(L_Matrix(L_Dimension, L_Dimension))
    return
  end subroutine Basis_Initialize

  function Combin_Num(N, M)
    implicit none
    integer(kind=8),intent(in):: N, M
    integer(kind=8):: Combin_Num
    integer(kind=4):: i
    real(kind=8):: C
    if(N>=M.and.N>0) then
      C = 1
      do i=1, M
        C = C * (N-M+i)/i
      end do
      Combin_Num = C
    else
      Combin_Num = 0
    end if
    return
  end function Combin_Num

  subroutine Next_State(State)
    implicit none
    integer(kind=8),intent(inout):: State 
    integer(kind=8):: State_D
    integer(kind=1):: i
    State_D = 0
    do i=0, 61
      if(btest(State, i)) then
        State = ibclr(State, i)
        if(.not.btest(State, i+1)) then
          State = ibset(State, i+1)
          exit
        end if
        State_D = ibset(ishft(State_D, 1), 0)
      end if
    end do
    State = State + State_D
    return 
  end subroutine Next_State

  function Trun_State(State)
    implicit none
    integer(kind=8),intent(in):: State
    logical:: Trun_State
    Trun_State = .true.
    if(Monopole_E(State)>Monopole_Trun) Trun_State = .false.
    return
  end function Trun_State

  subroutine Vector_Initialize()
    implicit none
    if(Print_Mode==0) then
      write(*, "(1X,A)") "Calculating..."
      write(*, "(3X,A)") "Calculation Starting..."
    end if
    !$OMP PARALLEL
    !$OMP WORKSHARE
    I_Vector = 0
    !$OMP END WORKSHARE NOWAIT
    !$OMP END PARALLEL
    I_Vector(Posit_Min) = 1
    Convergence = Run_Mode
    return 
  end subroutine Vector_Initialize

  subroutine Vector_Restart()
    implicit none
    if(Print_Mode==0) write(*, "(3X,A)") "Restarting..."
    !$OMP PARALLEL
    !$OMP WORKSHARE
    I_Vector = Q_Matrix(:, 1)
    !$OMP END WORKSHARE NOWAIT
    !$OMP END PARALLEL
    return 
  end subroutine Vector_Restart

  subroutine Bina_State(D, L)
    implicit none
    integer(kind=8),intent(in):: D, L
    integer(kind=8):: State_M, State_P, low, mid, hig
    integer(kind=1):: i, j 
    integer(kind=4):: TN
    TN = 1
    !$ TN = OMP_GET_THREAD_NUM() + 1
    i = 0
    j = 0
    do State_P=0, Total-1
      if(btest(B_Array(D), State_P)) then
        i = i + 1
        O_Array(i, TN) = State_P
      else
        j = j + 1
        V_Array(j, TN) = State_P
      end if
    end do
    do i=1, Occup
      State_M = ibclr(B_Array(D), O_Array(i, TN))
      do j=1, Total-Occup
        State_P = ibset(State_M, V_Array(j, TN))
        if(Trun_State(State_P)) then
          low = 1
          hig = B_Dimension
          do while(low<=hig)
            mid = (low + hig)/2
            if(B_Array(mid)==State_P) exit
            if(B_Array(mid)>State_P) hig = mid - 1
            if(B_Array(mid)<State_P) low = mid + 1
          end do
          I_Vector(D) = I_Vector(D) + P_F(O_Array(i, TN)+1, V_Array(j, TN)+1) * Q_Matrix(mid, L)
        end if
      end do
    end do
    return 
  end subroutine Bina_State

  subroutine Hash_State(D, L)
    implicit none
    integer(kind=8),intent(in):: D, L
    integer(kind=8):: State_M, State_P, low, mid, hig
    integer(kind=1):: i, j 
    integer(kind=4):: TN
    TN = 1
    !$ TN = OMP_GET_THREAD_NUM() + 1
    i = 0
    j = 0
    do State_P=0, Total-1
      if(btest(B_Array(D), State_P)) then
        i = i + 1
        O_Array(i, TN) = State_P
      else
        j = j + 1
        V_Array(j, TN) = State_P
      end if
    end do
    do i=1, Occup
      State_M = ibclr(B_Array(D), O_Array(i, TN))
      do j=1, Total-Occup
        State_P = ibset(State_M, V_Array(j, TN))
        mid = 1
        hig = 0
        do low=1, Total
          if(btest(State_P, low-1)) then
            hig = hig + 1
            mid = mid + C_Array(low, hig)
          end if
        end do
        I_Vector(D) = I_Vector(D) + P_F(O_Array(i, TN)+1, V_Array(j, TN)+1) * Q_Matrix(mid, L)
      end do
    end do
    return 
  end subroutine Hash_State

  subroutine Lanczos_Iteration()
    implicit none
    integer(kind=8):: i, l
    if(Print_Mode==0) then
      write(*, "(3X,A)") "Lanczos Decomposing..."
      write(*, "(3X,A11,$)") "Process :  "
    end if
    !$OMP PARALLEL
    !$OMP WORKSHARE
    L_Matrix = 0
    !$OMP END WORKSHARE NOWAIT
    !$OMP END PARALLEL
    do l=1, L_Dimension
      !$OMP PARALLEL
      !$OMP WORKSHARE
      Q_Matrix(:, l) = I_Vector
      !$OMP END WORKSHARE NOWAIT
      !$OMP END PARALLEL
      if(Truncated==0) then
        !$OMP PARALLEL PRIVATE(i)
        !$OMP DO SCHEDULE(DYNAMIC, 1)
        do i=1, B_Dimension
          I_Vector(i) = Monopole_E(B_Array(i)) * Q_Matrix(i, l)
          call Hash_State(i, l)
        end do
        !$OMP END DO NOWAIT
        !$OMP END PARALLEL
      else
        !$OMP PARALLEL PRIVATE(i)
        !$OMP DO SCHEDULE(DYNAMIC, 1)
        do i=1, B_Dimension
          I_Vector(i) = Monopole_E(B_Array(i)) * Q_Matrix(i, l)
          call Bina_State(i, l)
        end do
        !$OMP END DO NOWAIT
        !$OMP END PARALLEL
      end if
      !$OMP PARALLEL
      !$OMP WORKSHARE
      L_Matrix(l, l) = dot_product(I_Vector, Q_Matrix(:, l))
      !$OMP END WORKSHARE NOWAIT
      !$OMP END PARALLEL
      if(l/=L_Dimension) then
        !$OMP PARALLEL
        !$OMP WORKSHARE
        I_Vector = I_Vector - L_Matrix(l, l) * Q_Matrix(:, l)
        !$OMP END WORKSHARE NOWAIT
        !$OMP END PARALLEL
        if(l>1) then
          !$OMP PARALLEL
          !$OMP WORKSHARE
          I_Vector = I_Vector - L_Matrix(l, l-1) * Q_Matrix(:, l-1)
          !$OMP END WORKSHARE NOWAIT
          !$OMP END PARALLEL
        end if
        !$OMP PARALLEL
        !$OMP WORKSHARE
        Q_Matrix(:, L_Dimension) = 0
        !$OMP END WORKSHARE NOWAIT
        !$OMP END PARALLEL
        do i=1, l
          !$OMP PARALLEL
          !$OMP WORKSHARE
          Q_Matrix(:, L_Dimension) = Q_Matrix(:, L_Dimension) + dot_product(I_Vector, Q_Matrix(:, i)) * Q_Matrix(:, i)
          !$OMP END WORKSHARE NOWAIT
          !$OMP END PARALLEL
        end do
        !$OMP PARALLEL
        !$OMP WORKSHARE
        I_Vector = I_Vector - Q_Matrix(:, L_Dimension)
        L_Matrix(l, l+1) = sqrt(dot_product(I_Vector, I_Vector))
        I_Vector = I_Vector/L_Matrix(l, l+1)
        !$OMP END WORKSHARE NOWAIT
        !$OMP END PARALLEL
        L_Matrix(l+1, l) = L_Matrix(l, l+1)
        if(abs(L_Matrix(l+1, l)/L_Matrix(l, l))<=Lanc_Error) then
          L_Dimension = l
          Convergence = 1
          exit
        end if
      else
        if(L_Dimension==1) then
          Convergence = 1
          exit
        end if
      end if
      if(Print_Mode==0) write(*, "(A1,$)") "#"
    end do
    if(Print_Mode==0) write(*, "(3X,A8,F15.10)") "Ratio1 :", abs(L_Matrix(1, 2)/L_Matrix(1, 1))
    return 
  end subroutine Lanczos_Iteration

  subroutine QR_Decompose()
    implicit none
    integer(kind=8):: i, j, k, l
    real(kind=8):: drift, c, s, r
    real(kind=8),dimension(2,2):: G_Matrix
    if(Print_Mode==0) write(*, "(3X,A)") "QR Decomposing..."
    allocate(Q_Vector(L_Dimension))
    allocate(T_Matrix(L_Dimension, L_Dimension))
    allocate(P_Matrix(L_Dimension, L_Dimension))
    !$OMP PARALLEL PRIVATE(i, j)
    !$OMP DO SCHEDULE(DYNAMIC, 1)
    do i=1, L_Dimension
      do j=1, L_Dimension
        P_Matrix(i, j) = 0
      end do
      P_Matrix(i, i) = 1
    end do
    !$OMP END DO NOWAIT
    !$OMP END PARALLEL
    k = 0
    l = L_Dimension
    do while(l/=1)
      drift = L_Matrix(l, l)
      !$OMP PARALLEL PRIVATE(i, j)
      !$OMP DO SCHEDULE(DYNAMIC, 1)
      do i=1, l
        L_Matrix(i, i) = L_Matrix(i, i) - drift
      end do
      !$OMP END DO NOWAIT
      !$OMP DO SCHEDULE(DYNAMIC, 1)
      do i=1, L_Dimension
        do j=1, L_Dimension
          T_Matrix(i, j) = 0
        end do
        T_Matrix(i, i) = 1
      end do
      !$OMP END DO NOWAIT
      !$OMP END PARALLEL
      do i=1, l-1
        c = L_Matrix(i, i)
        s = L_Matrix(i+1, i)
        r = sqrt(c**2 + s**2)
        G_Matrix(1, 1) =  c/r
        G_Matrix(2, 2) =  c/r
        G_Matrix(2, 1) = -s/r
        G_Matrix(1, 2) =  s/r
        j = 2
        if(i==l-1) j = 1
        L_Matrix(i:i+1, i:i+j) = matmul(G_Matrix, L_Matrix(i:i+1, i:i+j))
        G_Matrix(2, 1) =  s/r
        G_Matrix(1, 2) = -s/r
        T_Matrix(1:i+1, i:i+1) = matmul(T_Matrix(1:i+1, i:i+1), G_Matrix)
      end do
      !$OMP PARALLEL PRIVATE(i, j, Q_Vector)
      !$OMP DO SCHEDULE(DYNAMIC, 1)
      do i=1, L_Dimension
        Q_Vector(1:l) = L_Matrix(i, 1:l)
        do j=1, l
          L_Matrix(i, j) = dot_product(Q_Vector(1:l), T_Matrix(1:l, j))
          if(j==i) L_Matrix(i, j) = L_Matrix(i, j) + drift
        end do
      end do
      !$OMP END DO NOWAIT
      !$OMP DO SCHEDULE(DYNAMIC, 1)
      do i=1, L_Dimension
        Q_Vector(1:l) = P_Matrix(i, 1:l)
        do j=1, l
          P_Matrix(i, j) = dot_product(Q_Vector(1:l), T_Matrix(1:l, j))
        end do
      end do
      !$OMP END DO NOWAIT
      !$OMP END PARALLEL
      k = k + 1
      if((abs(L_Matrix(l,l-1))/(abs(L_Matrix(l-1,l-1))+abs(L_Matrix(l,l)))<1D-10).or.(k>100)) then
        l = l - 1
        k = 0
      end if
    end do
    do i=1, L_Dimension-1
      s = L_Matrix(i, i)
      k = i
      do j=i+1, L_Dimension
        if(s>L_Matrix(j, j)) then
          k = j
          s = L_Matrix(j, j)
        end if
      end do
      L_Matrix(k, k) = L_Matrix(i, i)
      L_Matrix(i, i) = s
      Q_Vector = P_Matrix(:, i)
      P_Matrix(:, i) = P_Matrix(:, k)
      P_Matrix(:, k) = Q_Vector 
    end do
    !$OMP PARALLEL PRIVATE(i, j, Q_Vector)
    !$OMP DO SCHEDULE(DYNAMIC, 1)
    do i=1, B_Dimension
      Q_Vector(1:L_Dimension) = Q_Matrix(i, 1:L_Dimension)
      do j=1, L_Dimension
        Q_Matrix(i, j) = dot_product(Q_Vector(1:L_Dimension), P_Matrix(1:L_Dimension, j))
      end do
    end do
    !$OMP END DO NOWAIT
    !$OMP END PARALLEL
    deallocate(Q_Vector)
    deallocate(T_Matrix)
    deallocate(P_Matrix)
    return 
  end subroutine QR_Decompose

  subroutine Results_Output()
    implicit none
    integer(kind=8):: i, j
    real(kind=8):: Percent
    N_Occup = 0
    !$OMP PARALLEL PRIVATE(i, j, Percent) SHARED(N_Occup)
    !$OMP DO SCHEDULE(DYNAMIC, 1) REDUCTION(+:N_Occup)
    do i=1, B_Dimension
      Percent = Q_Matrix(i, 1) * Q_Matrix(i, 1)
      do j=1, Total
        if(btest(B_Array(i), j-1)) then
          N_Occup(j) = N_Occup(j) + 2 * Percent
        end if
      end do
    end do
    !$OMP END DO NOWAIT
    !$OMP END PARALLEL
    Energy_Ground = L_Matrix(1, 1)
    if(Run_Mode/=0) then
      do i=1, min(5, L_Dimension)
        Eigens(i) = L_Matrix(i, i)
      end do
    end if
    if(Print_Mode==0) then
      write(*, *)
      write(*, "(1X,A)") "Results..."
      write(*, "(3X,A17,F22.14)") "Energy_Ground   :", Energy_Ground
      write(*, "(3X,A17,F22.14)") "Monopole_Min    :", Monopole_Min
      write(*, *)
      if(Run_Mode/=0) then
        write(*, "(1X,A)") "First Five Eigenvalues..."
        do i=1, min(5, L_Dimension)
          write(*, "(3X,A11,I1,A3,F22.14)") "Eigenvalue ", i, " : ", Eigens(i)
        end do
        write(*, *)
      end if
    end if
    return
  end subroutine Results_Output

  subroutine Destory()
    implicit none
    deallocate(B_Array)
    deallocate(O_Array)
    deallocate(V_Array)
    deallocate(C_Array)
    deallocate(I_Vector)
    deallocate(Q_Matrix)
    deallocate(L_Matrix)
    return
  end subroutine Destory

end module PairDiag

